import functools
from typing import Generator

import torch
import torch.autograd as autograd
from torch import Tensor
from torch.optim import Optimizer


def get_parameters(optimizer: Optimizer) -> Generator[Tensor, None, None]:
    yield from (
        parameter
        for parameter_group in optimizer.param_groups
        for _, parameters in parameter_group.items()
        for parameter in (parameters if isinstance(parameters, list) else [parameters])
        if torch.is_tensor(parameter) and parameter.requires_grad
    )


def manual_step(optimizer: Optimizer, loss: Tensor, allow_unused: bool = False) -> None:
    optimizer.zero_grad()

    parameters = list(get_parameters(optimizer))
    gradients = autograd.grad(loss, parameters, retain_graph=True, allow_unused=allow_unused)

    for parameter, gradient in zip(parameters, gradients):
        parameter.grad = gradient

    optimizer.step()
    optimizer.zero_grad()


def adaptive_clipping(optimizer: Optimizer, loss: Tensor, target_loss: Tensor, scale_factor: float) -> None:
    optimizer.zero_grad()

    parameters = list(get_parameters(optimizer))
    loss_gradients = autograd.grad(loss, parameters, retain_graph=True, allow_unused=True)
    target_gradients = autograd.grad(target_loss, parameters, retain_graph=True, allow_unused=True)

    target_norm, loss_norm = functools.reduce(
        lambda tuple_first, tuple_second: (tuple_first[0] + tuple_second[0], tuple_first[1] + tuple_second[1]),
        (
            (torch.sum(target_gradient ** 2), torch.sum(loss_gradient ** 2))
            for target_gradient, loss_gradient in zip(target_gradients, loss_gradients)
            if target_gradient is not None and loss_gradient is not None
        ),
    )
    target_norm, loss_norm = torch.sqrt(target_norm), torch.sqrt(loss_norm)

    scaling_factor = torch.minimum(loss_norm, target_norm) / loss_norm * scale_factor

    for parameter, target_gradient, loss_gradient in zip(parameters, target_gradients, loss_gradients):
        if target_gradient is not None:
            parameter.grad = target_gradient

        if loss_gradient is not None:
            if parameter.grad is not None:
                parameter.grad += scaling_factor * loss_gradient
            else:
                parameter.grad = loss_gradient

    optimizer.step()
    optimizer.zero_grad()



import logging
import os
from typing import Optional, Dict, Any

import mlflow
import pytorch_lightning.loggers as loggers
import pytorch_lightning.loggers.base as base
from mlflow.tracking import MlflowClient
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME
from pytorch_lightning import Trainer, LightningModule
from pytorch_lightning.callbacks import ModelCheckpoint

import sfu_torch_lib.io as io


log = logging.getLogger(__name__)


class MLFlowLogger(loggers.MLFlowLogger):
    def __init__(
        self,
        run_name: Optional[str] = None,
        tracking_uri: Optional[str] = os.getenv('MLFLOW_TRACKING_URI'),
        tags: Optional[Dict[str, Any]] = None,
        save_dir: Optional[str] = './mlruns',
        prefix: str = "",
        artifact_location: Optional[str] = None,
    ) -> None:

        super().__init__(
            run_name=run_name,
            tracking_uri=tracking_uri,
            tags=tags,
            save_dir=save_dir,
            prefix=prefix,
            artifact_location=artifact_location,
        )

    @property  # type: ignore
    @base.rank_zero_experiment
    def experiment(self) -> MlflowClient:
        if self._run_id and self._experiment_id:
            return self._mlflow_client

        self.tags = self.tags or {}

        if self._run_name is not None:
            if MLFLOW_RUN_NAME in self.tags:
                log.warning(
                    f'The tag {MLFLOW_RUN_NAME} is found in tags. '
                    f'The value will be overridden by {self._run_name}.'
                )

            self.tags[MLFLOW_RUN_NAME] = self._run_name

        run = mlflow.active_run() if mlflow.active_run() else mlflow.start_run(tags=self.tags)

        self._run_id = run.info.run_id
        self._experiment_id = run.info.experiment_id
        self._experiment_name = mlflow.get_experiment(run.info.experiment_id).name

        return self._mlflow_client


class MLFlowModelCheckpoint(ModelCheckpoint):
    def __init__(self, patience: int = 0, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.patience = patience
        self.mlflow_epoch = 0
        self.mlflow_model_path = ''

    @staticmethod
    def log_artifact(local_file: str, filename: str = 'last.ckpt') -> None:
        bucket, key = io.get_bucket_and_key(mlflow.get_artifact_uri())
        destination_path = os.path.join(key, filename)

        io.upload_s3(bucket, destination_path, local_file)

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_train_epoch_end(trainer, pl_module)

        should_save = trainer.current_epoch - self.mlflow_epoch >= self.patience
        has_new_model = self.best_model_path and self.best_model_path != self.mlflow_model_path

        if should_save and has_new_model:
            self.log_artifact(self.best_model_path)

            self.mlflow_epoch = trainer.current_epoch
            self.mlflow_model_path = self.best_model_path

    def on_train_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_train_end(trainer, pl_module)

        self.log_artifact(self.best_model_path)



import functools
import math
import re
from typing import List, Tuple, TypeVar, Generator, Type, Union, Sequence, Mapping, Callable

from torch.nn import ELU, Identity, ReLU, LeakyReLU, Sigmoid, Softmax, Tanh, Module
from torch.utils.data import Dataset


T = TypeVar('T')


def get_pairs(items: List[T]) -> Generator[Tuple[T, T], None, None]:
    for index in range(len(items) - 1):
        yield items[index], items[index + 1]


def create_selector(indices: Union[int, Sequence[int]]) -> Callable[[Sequence[T]], Union[T, Tuple[T, ...]]]:
    def select(items: Sequence[T]) -> Union[T, Tuple[T, ...]]:
        if isinstance(indices, int):
            return items[indices]

        return tuple(items[index] for index in indices)

    return select


def to_list(*args: Union[Sequence[T], T]) -> List[T]:
    items: List[T] = []

    for argument in args:
        if isinstance(argument, Sequence):
            items += argument
        else:
            items.append(argument)

    return items


def get_activation(name: str) -> Type[Module]:
    return {
        'elu': ELU,
        'identity': Identity,
        'relu': ReLU,
        'leaky_relu': LeakyReLU,
        'sigmoid': Sigmoid,
        'softmax': Softmax,
        'tanh': Tanh,
    }[name]


def parse_map_sequence(string: str, sequence_type: Type[T]) -> Mapping[str, Sequence[T]]:
    substrings = re.split(r'(?<=]),', string)

    output = {}

    for substring in substrings:
        key, value = substring.split(':')

        output[key] = list(map(sequence_type, value.strip('[]').split(',')))

    return output


def get_num_steps(dataset: Dataset, batch_size: int, frequency: int = 10) -> int:
    return max(1, math.floor(len(dataset) / batch_size / frequency))


parse_map_ints = functools.partial(parse_map_sequence, sequence_type=int)
parse_map_floats = functools.partial(parse_map_sequence, sequence_type=float)
parse_map_strings = functools.partial(parse_map_sequence, sequence_type=str)
